# -*- coding: utf-8 -*-

# Copyright (c) 2014, Joshua Stone
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from .request import get_user_data, get_user_id_by_name
from .tag import Tag
from .comments import Comments
from .search import Search

__all__ = [
  "Profile"
]

class Profile(object):
  def __init__(self, user_id, username="", proxies={}):
    self.proxies = proxies
    if user_id is None:
      user_id = get_user_id_by_name(username, proxies=proxies)
    self._data = get_user_data(user_id, proxies=proxies)
    for field, body in self.data.items():
      if not hasattr(self, field):
        setattr(self, field, body)

  def __str__(self):
    return f'''Profile({self.name})'''

  @property
  def data(self):
    return self._data
       
  @property
  def url(self):
    return f"https://derpibooru.org/profiles/{self.slug}"

  def update(self):
    data = get_user_data(self.id, proxies=self.proxies)

    if data:
      self._data = data

  def awards(self):
    for award_data in self.data['awards']:
      yield Award(award_data)

  def links(self):
    for link_data in self.data['links']:
      yield Link(link_data, proxies=self.proxies)

  def comments(self, key="", limit=50, filter_id="", per_page=25, page=1):
    return Comments(user_id=self.id, key=key, limit=limit, filter_id=filter_id,
                    per_page=per_page, page=page, proxies=self.proxies)

  def uploads(self, key="", sf="created_at", sd="desc", limit=50,
              filter_id="", per_page=25, page=1):
    return Search(q=(f"uploader_id:{self.id}",), key=key, sf=sf, sd=sd,
                  limit=limit, filter_id=filter_id, per_page=per_page,
                  page=page, proxies=self.proxies)

  def favorites(self, key="", sf="created_at", sd="desc", limit=50,
                filter_id="", per_page=25, page=1):
    return Search(q=(f"faved_by_id:{self.id}",), key=key, sf=sf, sd=sd,
                  limit=limit, filter_id=filter_id, per_page=per_page,
                  page=page, proxies=self.proxies)

  def artworks(self, key="", sf="created_at", sd="desc", limit=50,
               filter_id="", per_page=25, page=1):
    artist_tags = {link.tag.name for link in self.links()}
    return Search(q=artist_tags, key=key, sf=sf, sd=sd,
                  limit=limit, filter_id=filter_id, per_page=per_page,
                  page=page, proxies=self.proxies)

class Award(object):
  def __init__(self, data):
    self._data = data
    for field, body in self.data.items():
      if not hasattr(self, field):
        setattr(self, field, body)

  def __str__(self):
    return f'''Badge({self.title})'''

  @property
  def data(self):
    return self._data

class Link(object):
  def __init__(self, data, proxies={}):
    self.proxies = proxies
    self._data = data
    for field, body in self.data.items():
      if not hasattr(self, field):
        setattr(self, field, body)

  def __str__(self):
    return f'''Link({self.title})'''

  @property
  def data(self):
    return self._data

  @property
  def tag(self):
    return Tag(None, tag_id=self.data["tag_id"], proxies=self.proxies)




# -*- coding: utf-8 -*-

# Copyright (c) 2014, Joshua Stone
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from .request import get_tag_data, get_tags
from .helpers import slugging_tag

__all__ = [
  "Tag"
]

class Tag(object):
  def __init__(self, data, tag=None, slug=False, tag_id=None, proxies={}):
    """
    tag field is slug.
    """
    self.proxies = proxies
    if data is None and (tag or tag_id):
      if tag:
        if slug:
          self._data = data = get_tag_data(tag, proxies=proxies)
        else:
          self._data = data = get_tag_data(slugging_tag(tag), proxies=proxies)
      elif tag_id:
          self._data = data = next(get_tags({"q": (f"id:{tag_id}",), "per_page":1}, 1, proxies=proxies))
    else:
      self._data = data
    for field, body in self.data.items():
      if not hasattr(self, field):
        setattr(self, field, body)

  def __str__(self):
    return f'''Tag({self.name})'''

  @property
  def data(self):
    return self._data
       
  @property
  def url(self):
    """
    Also url for search image by single tag.
    """
    return f"https://derpibooru.org/tags/{self.slug}"

  def update(self):
    data = get_tag_data(self.id, proxies=self.proxies)

    if data:
      self._data = data

  def allias_parent(self):
    """
    Return main Tag() for this tag.
    """
    if self.aliased_tag:
      return Tag(None, tag=self.aliased_tag, slug=True, proxies=self.proxies)

  def allias_children(self):
    """
    Return in generator all alliases tags.
    """
    for tag in self.aliases:
      yield Tag(None, tag=tag, slug=True, proxies=self.proxies)

  def implied(self):
    for tag in self.implied_tags:
      yield Tag(None, tag=tag, slug=True, proxies=self.proxies)

  def implied_by(self):
    for tag in self.implied_by_tags:
      yield Tag(None, tag=tag, slug=True, proxies=self.proxies)


from derpibooru import Search

def test_query():
  """
  Tests whether the results in a query contain the tag that was being searched
  for
  """
  limit, tag = 10, "sunset shimmer"
  images = [image for image in Search().query(tag).limit(limit)]

  assert len(images) == limit

  for image in images:
    assert tag in image.tags

def test_ascending():
  """
  Tests whether ascending search is in the correct order
  """
  limit = 10
  images = [image for image in Search().ascending().limit(limit)]

  assert len(images) == limit

  for image in images:
    # Check if the images are in ascending order
    # by comparing the ID of the next image 
    if image is not images[-1]:
      next_image = images[images.index(image) + 1]
      assert image.id_number < next_image.id_number

def test_descending():
  """
  Tests whether descending search is in the correct order
  """
  limit = 10
  images = [image for image in Search().descending().limit(limit)]

  assert len(images) == limit

  for image in images:
  # Check if the image IDs are listed in descending order
  # by comparing the ID of the next image
    if image is not images[-1]:
      next_image = images[images.index(image) + 1]
      assert image.id_number > next_image.id_number




# -*- coding: utf-8 -*-

# Copyright (c) 2014, Joshua Stone
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from .request import get_filters, get_filter_data
from .helpers import api_key, join_params, set_limit, validate_filter

__all__ = [
  "Filters",
  "Filter"
]

class Filters(object):
  def __init__(self, key="", filters_id="", limit=50, per_page=25, page=1, proxies={}):
    self.proxies = proxies
    self._params = {
      "key": api_key(key),
      "per_page": set_limit(per_page),
      "page": set_limit(page)
    }

    if filters_id not in {"system", "user"}:
      if key:
        filters_id = "user"
      else:
        filters_id = "system"
      
    self._limit = set_limit(limit)
    self._search = get_filters(filters_id, self._params, self._limit, proxies=self.proxies)
  
  def __iter__(self):
    """
    Make Filters() iterable so that new search results can be lazily generated
    for performance reasons.
    """
    return self

  @property
  def parameters(self):
    """
    Returns a list of available parameters; useful for passing state to new
    instances of Filters().
    """
    return self._params

  @property
  def url(self):
    """
    Returns a standart URL of avaliable filters list
    """
    return "https://derpibooru.org/filters"

  def key(self, key=""):
    """
    Takes a user's API key string which applies content settings. API keys can
    be found at <https://derpibooru.org/registration/edit>.
    """
    params = join_params(self.parameters, {"key": key, "proxies": self.proxies})

    return self.__class__(**params)

  def limit(self, limit):
    """
    Set absolute limit on number of filters to return, or set to None to return
    as many results as needed; default 50 posts. This limit on app-level.
    """
    self._limit = set_limit(limit)

    return self.__class__(**params)

  def get_page(self,page):
    """
    Set page for gets result of search.
    """
    params = join_params(self.parameters, {"page": set_limit(page), "proxies": self.proxies})

    return self.__class__(**params)

  def per_page(self,limit):
    """
    Set absolute limit on number of filters to get, or set to None to return
    defaulting 25 posts; max 50 posts. This limit on API-level.
    """
    params = join_params(self.parameters, {"per_page": set_limit(limit), "proxies": self.proxies})

    return self.__class__(**params)

  def __next__(self):
    """
    Returns a result wrapped in a new instance of Filter().
    """
    return Filter(None, data=next(self._search), proxies=self.proxies)

class Filter(object):
  """
  This class provides a thin wrapper around JSON data, mapping each value to
  its own property. Once instantiated the data is immutable so as to reflect
  the stateless nature of a REST API.
  """
  def __init__(self, filter_id, data=None, proxies={}):
    self.proxies = proxies

    if filter_id is None and data:
      self._data = data
    else:
      norm_str_filter_id = f"{filter_id}".lower()
      if norm_str_filter_id=="default":
        filter_id = 100073
      elif norm_str_filter_id=="everything":
        filter_id = 56027
      elif norm_str_filter_id=="r34":
        filter_id = 37432
      elif norm_str_filter_id=="legacy default":
        filter_id = 37431
      elif norm_str_filter_id=="maximum spoilers":
        filter_id = 37430
      elif norm_str_filter_id=="dark":
        filter_id = 37429
      self._data = data = get_filter_data(validate_filter(filter_id), proxies=proxies)

    for field, body in data.items():
      if not hasattr(self, field):
        setattr(self, field, body) 

  def __str__(self):
    return f"Filter({self.name})"

  @property
  def data(self):
    return self._data

  def update(self):
    data = get_image_data(self.id, proxies=self.proxies)

    if data:
      self._data = data


# -*- coding: utf-8 -*-

# Copyright (c) 2014, Joshua Stone
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from .request import get_images, url, get_related, url_related
from .image import Image
from .helpers import tags, api_key, sort_format, join_params, user_option, set_limit, \
                     validate_filter, set_distance

__all__ = [
  "Search"
]

class Search(object):
  """
  Search() is the primary interface for interacting with Derpibooru's REST API.

  All properties are read-only, and every method returns a new instance of
  Search() to avoid mutating state in ongoing search queries. This makes object
  interactions predictable as well as making versioning of searches relatively
  easy.
  """
  def __init__(self, key="", q=set(), sf="created_at", sd="desc",
               limit=50, faves="", upvotes="", uploads="", watched="",
               filter_id="", per_page=25, page=1,
               reverse_url="", distance=0.25, proxies={}):
    """
    By default initializes an instance of Search with the parameters to get
    the first 25 images on Derpibooru's front page.
    For reverse searching by image use reverse_url field.
    """
    self.proxies = proxies
    self._params = {
      "key": api_key(key),
      "reverse_url": reverse_url,
      "distance": set_distance(distance),
      "q": tags(q),
      "sf": sort_format(sf),
      "sd": sd,
      "filter_id": validate_filter(filter_id),
      "per_page": set_limit(per_page),
      "page": set_limit(page)
    }

    if faves:
      if self._params["key"] and faves is user.ONLY:
         self._params["q"] -= {"-my:faves"}
         self._params["q"].add("my:faves")
      elif self._params["key"] and faves is user.NOT:
         self._params["q"] -= {"my:faves"}
         self._params["q"].add("-my:faves")
    if upvotes:
      if self._params["key"] and upvotes is user.ONLY:
         self._params["q"] -= {"-my:upvotes"}
         self._params["q"].add("my:upvotes")
      elif self._params["key"] and upvotes is user.NOT:
         self._params["q"] -= {"my:upvotes"}
         self._params["q"].add("-my:upvotes")
    if uploads:
      if self._params["key"] and uploads is user.ONLY:
         self._params["q"] -= {"-my:uploads"}
         self._params["q"].add("my:uploads")
      elif self._params["key"] and uploads is user.NOT:
         self._params["q"] -= {"my:uploads"}
         self._params["q"].add("-my:uploads")
    if watched:
      if self._params["key"] and watched is user.ONLY:
         self._params["q"] -= {"-my:watched"}
         self._params["q"].add("my:watched")
      elif self._params["key"] and watched is user.NOT:
         self._params["q"] -= {"my:watched"}
         self._params["q"].add("-my:watched")
      
    self._limit = set_limit(limit)
    self._search = get_images(self._params, self._limit, proxies=self.proxies)
  
  def __iter__(self):
    """
    Make Search() iterable so that new search results can be lazily generated
    for performance reasons.
    """
    return self

  @property
  def parameters(self):
    """
    Returns a list of available parameters; useful for passing state to new
    instances of Search().
    """
    return self._params

  @property
  def url(self):
    """
    Returns a search URL built on set parameters. Example based on default
    parameters:

    https://derpibooru.org/search?sd=desc&sf=created_at&q=%2A
    """
    return url(self.parameters)

  def key(self, key=""):
    """
    Takes a user's API key string which applies content settings. API keys can
    be found at <https://derpibooru.org/registration/edit>.
    """
    params = join_params(self.parameters, {"key": key, "limit": self._limit, "proxies": self.proxies})

    return self.__class__(**params)

  def query(self, *q):
    """
    Takes one or more strings for searching by tag and/or metadata.
    """
    params = join_params(self.parameters, {"q": q, "limit": self._limit, "proxies": self.proxies})

    return self.__class__(**params)

  def sort_by(self, sf):
    """
    Determines how to sort search results. Available sorting methods are
    sort.SCORE, sort.COMMENTS, sort.HEIGHT, sort.RELEVANCE, sort.CREATED_AT,
    and sort.RANDOM; default is sort.CREATED_AT.
    """
    params = join_params(self.parameters, {"sf": sf, "limit": self._limit, "proxies": self.proxies})

    return self.__class__(**params)

  def descending(self):
    """
    Order results from largest to smallest; default is descending order.
    """
    params = join_params(self.parameters, {"sd": "desc", "limit": self._limit, "proxies": self.proxies})

    return self.__class__(**params)

  def ascending(self, sd="asc"):
    """
    Order results from smallest to largest; default is descending order.
    """
    params = join_params(self.parameters, {"sd": sd, "limit": self._limit, "proxies": self.proxies})

    return self.__class__(**params)

  def limit(self, limit):
    """
    Set absolute limit on number of images to return, or set to None to return
    as many results as needed; default 50 posts. This limit on app-level.
    """
    params = join_params(self.parameters, {"limit": limit, "proxies": self.proxies})

    return self.__class__(**params)

  def filter(self, filter_id=""):
    """
    Takes a filter's ID to be used in the current search context. Filter IDs can
    be found at <https://derpibooru.org/filters/> by inspecting the URL parameters.
    
    If no filter is provided, the user's current filter will be used.
    """
    params = join_params(self.parameters, {"filter_id": validate_filter(filter_id), "limit": self._limit, "proxies": self.proxies})

    return self.__class__(**params)


  def faves(self, option):
    """
    Set whether to filter by a user's faves list. Options available are
    user.ONLY, user.NOT, and None; default is None.
    """
    if self.parameters["key"] and option is user.ONLY:
       query = self.query_remove("-my:faves").query_append("my:faves")
    elif self.parameters["key"] and option is user.NOT:
       query = self.query_remove("my:faves").query_append("-my:faves")
    else:
       query = self.query_remove("my:faves").query_remove("-my:faves")
    return query

  def upvotes(self, option):
    """
    Set whether to filter by a user's upvoted list. Options available are
    user.ONLY, user.NOT, and None; default is None.
    """
    if self.parameters["key"] and option is user.ONLY:
       query = self.query_remove("-my:upvotes").query_append("my:upvotes")
    elif self.parameters["key"] and option is user.NOT:
       query = self.query_remove("my:upvotes").query_append("-my:upvotes")
    else:
       query = self.query_remove("my:upvotes").query_remove("-my:upvotes")
    return query

  def uploads(self, option):
    """
    Set whether to filter by a user's uploads list. Options available are
    user.ONLY, user.NOT, and None; default is None.
    """
    if self.parameters["key"] and option is user.ONLY:
       query = self.query_remove("-my:uploads").query_append("my:uploads")
    elif self.parameters["key"] and option is user.NOT:
       query = self.query_remove("my:uploads").query_append("-my:uploads")
    else:
       query = self.query_remove("my:uploads").query_remove("-my:uploads")
    return query

  def watched(self, option):
    """
    Set whether to filter by a user's watchlist. Options available are
    user.ONLY, user.NOT, and None; default is None.
    """
    if self.parameters["key"] and option is user.ONLY:
       query = self.query_remove("-my:watched").query_append("my:watched")
    elif self.parameters["key"] and option is user.NOT:
       query = self.query_remove("my:watched").query_append("-my:watched")
    else:
       query = self.query_remove("my:watched").query_remove("-my:watched")
    return query

  def top(self):
     """
     Returns search for Trending Images from front page.
     """
     return self.query('first_seen_at.gt:3 days ago').sort_by(sort.SCORE)
  
  # ids - list of int
  def exclude_by_id(self,*ids):
     """
     Excludes images from search by id.
     """
     return self.query_remove(f"id:{elem}" for elem in ids).query_append(f"-id:{elem}" for elem in ids)

  def query_append(self,*q):
     """
     Adds tags to current search.
     """
     query = self.parameters['q'].union(q)
     params = join_params(self.parameters, {"q": query, "limit": self._limit, "proxies": self.proxies})

     return self.__class__(**params)

  def query_remove(self,*q):
     """
     Removes tags from current search.
     """
     query = self.parameters['q'].difference(q)
     params = join_params(self.parameters, {"q": query, "limit": self._limit, "proxies": self.proxies})

     return self.__class__(**params)

  def get_related(self,image):
    if isinstance(image,Image):
      return Related(image, key=self.parameters['key'], limit=self._limit, \
                     filter_id=self.parameters['filter_id'], per_page=self.parameters['per_page'], \
                     proxies=self.proxies)
    else:
      return Related(Image(None, image_id=image, proxies=self.proxies), \
                     key=self.parameters['key'], limit=self._limit, \
                     filter_id=self.parameters['filter_id'], per_page=self.parameters['per_page'], \
                     proxies=self.proxies)

  def get_page(self,page):
    """
    Set page for gets result of search.
    """
    params = join_params(self.parameters, {"page": set_limit(page), "limit": self._limit, "proxies": self.proxies})

    return self.__class__(**params)

  def per_page(self,limit):
    """
    Set absolute limit on number of images to get, or set to None to return
    defaulting 25 posts; max 50 posts. This limit on API-level.
    """
    params = join_params(self.parameters, {"per_page": set_limit(limit), "limit": self._limit, "proxies": self.proxies})

    return self.__class__(**params)

  def reverse(self,url):
    """
    Takes an url image for reverse search.
    """
    params = join_params(self.parameters, {"reverse_url": url, "limit": self._limit, "proxies": self.proxies})

    return self.__class__(**params)

  def distance(self,distance):
    """
    Match distance for reverse search (suggested values: between 0.2 and 0.5)
    """
    params = join_params(self.parameters, {"distance": set_distance(distance), "limit": self._limit, "proxies": self.proxies})

    return self.__class__(**params)

  def __next__(self):
    """
    Returns a result wrapped in a new instance of Image().
    """
    return Image(next(self._search), search_params=self.parameters, proxies=self.proxies)

class Related(Search):
  """
  Related() is the Search-like interface based on related images instead query.
  Related images gets with old API in JSON and with new API in Web.
  This class should returns Search() for any query-like actions.
  """
  def __init__(self, image, key="", limit=50,
               filter_id="", per_page=25, proxies={}):
    """
    By default initializes with the parameters to get the first 25 related images.
    """
    if proxies:
      self.proxies = proxies
    else:
      self.proxies = image.proxies
    self.image = image
    self._params = {
      "key": api_key(key) if key else api_key(image._params['key']),
      "filter_id": validate_filter(filter_id),
      "perpage": set_limit(per_page)
    }
    self._limit = set_limit(limit)
    self._search = get_related(self.image.id, self._params, self._limit, proxies=self.proxies)

  @property
  def url(self):
    """
    Returns a search URL built on set parameters. Example based on default
    parameters:

    https://derpibooru.org/images/***/related?key=&filter_id=&perpage=25
    """
    return url_related(self.image.id, self.parameters)

  def query(self, *q):
    """
    Takes one or more strings for searching by tag and/or metadata.
    """
    params = join_params(self.parameters, {"q": q, "proxies": self.proxies})
    return Search(**params)

  def sort_by(self, sf):
    """
    Related() can't be sorted.
    """
    return self

  def descending(self):
    """
    Related() can't be sorted.
    """
    return self

  def ascending(self, sd="asc"):
    """
    Related() can't be sorted.
    """
    return self
  
  def query_append(self,*q):
    """
    Synonyme of query() for Related().
    """
    return self.query(q)

  def query_remove(self,*q):
    """
    Nothing remove from Related().
    """
    return self

  def get_page(self,page):
    """
    Related() hasn't pages.
    """
    return self

  def per_page(self,limit):
    """
    Unlike Search(), Related() use perpage instead per_page.
    """
    params = join_params(self.parameters, {"perpage": set_limit(limit), "proxies": self.proxies})

    return self.__class__(**params)

  def reverse(self,url):
    """
    Takes an url image for reverse search.
    """
    params = join_params(self.parameters, {"reverse_url": url, "proxies": self.proxies})

    return Search(**params)

  def distance(self,distance):
    """
    It hasn't any sense in Related()
    """
    return self


# -*- coding: utf-8 -*-

# Copyright (c) 2014, Joshua Stone
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from requests import get, post, codes
from urllib.parse import urlencode
from .helpers import format_params

__all__ = [
  "url",
  "request",
  "get_images",
  "get_image_data",
  "set_limit"
]

def url(params):
  p = format_params(params)
  url = f"https://derpibooru.org/search?{urlencode(p)}"

  return url

def request(params, proxies={}):
  if "reverse_url" in params and params["reverse_url"]:
    search, p = "https://derpibooru.org/api/v1/json/search/reverse", format_params(params)
    p = {i:p[i] for i in p if i in ('url','distance')}
    request = post(search, params=p, proxies=proxies)
    p["per_page"] = 50
  else:
    search, p = "https://derpibooru.org/api/v1/json/search/images", format_params(params)
    p = {i:p[i] for i in p if i not in ('url','distance')}
    request = get(search, params=p, proxies=proxies)

  while request.status_code == codes.ok:
    images, image_count = request.json()["images"], 0
    for image in images:
      yield image
      image_count += 1
    if image_count < p["per_page"]:
      break

    p["page"] += 1

    request = get(search, params=p, proxies=proxies)

def get_images(params, limit=50, proxies={}):
  if limit is not None:
    if limit > 0:
      r = request(params, proxies=proxies)
      for index, image in enumerate(r, start=1):
        yield image
        if index >= limit:
          break
  else:
    r = request(params, proxies=proxies)
    for image in r:
      yield image

def get_image_data(id_number, proxies={}):
  """id_number can be featured"""
  url = f"https://derpibooru.org/api/v1/json/images/{id_number}"

  request = get(url, proxies=proxies)

  if request.status_code == codes.ok:
    data = request.json()

    if "duplicate_of" in data:
      return get_image_data(data["duplicate_of"], proxies=proxies)
    else:
      return data["image"]

def get_image_faves(id_number, proxies={}):
  url = f"https://derpibooru.org/images/{id_number}/favorites"

  request = get(url, proxies=proxies)

  if request.status_code == codes.ok:
    data = request.text.rsplit('</h5>',1)[-1].strip()
    if data.endswith('</a>'):
       data = data[:-4]
    data = data.split("</a> <")
    data = [useritem.rsplit('">',1)[-1] for useritem in data]
    return data

def url_related(id_number, params):
  p = format_params(params)
  url = f"https://derpibooru.org/images/{id_number}/related?{urlencode(p)}"

  return url

def request_related(id_number, params, proxies={}):
  search, p = f"https://www.derpibooru.org/images/{id_number}/related.json", format_params(params)
  request = get(search, params=p, proxies=proxies)

  while request.status_code == codes.ok:
    images, image_count = request.json()["images"], 0
    for image in images:
      image['view_url'] = image['image']
      yield image
      image_count += 1
    if image_count < p["per_page"]:
      break

    p["page"] += 1

    request = get(search, params=p, proxies=proxies)

def get_related(id_number, params, limit=50, proxies={}):
  if limit is not None:
    if limit > 0:
      r = request_related(id_number, params, proxies=proxies)
      for index, image in enumerate(r, start=1):
        yield image
        if index >= limit:
          break
  else:
    r = request_related(id_number, params, proxies=proxies)
    for image in r:
      yield image

def url_comments(params):
  p = format_params(params)
  p["qc"]=p["q"]
  del(p["q"])
  url = f"https://derpibooru.org/comments?{urlencode(p)}"

  return url

def comments_requests(params, proxies={}):
  search, p = "https://derpibooru.org/api/v1/json/search/comments", format_params(params)
  request = get(search, params=p, proxies=proxies)

  while request.status_code == codes.ok:
    comments, comment_count = request.json()["comments"], 0
    for comment in comments:
      yield comment
      comment_count += 1
    if comment_count < p["per_page"]:
      break

    p["page"] += 1

    request = get(search, params=p, proxies=proxies)

def get_comments(parameters, limit=50, proxies={}):
  params = parameters

  if limit is not None:
    if limit > 0:
      r = comments_requests(params, proxies=proxies)
      for index, comment in enumerate(r, start=1):
        yield comment
        if index >= limit:
          break
  else:
    r = comments_requests(params, proxies=proxies)
    for comment in r:
      yield comment

def get_comment_data(id_number, proxies={}):
  url = f"https://derpibooru.org/api/v1/json/comments/{id_number}"

  request = get(url, proxies=proxies)

  if request.status_code == codes.ok:
    data = request.json()

    return data["comment"]

def url_tags(params):
  p = format_params(params)
  p["tq"]=p["q"]
  del(p["q"])
  url = f"https://derpibooru.org/tags?{urlencode(p)}"

  return url

def tags_requests(params, proxies={}):
  search, p = "https://derpibooru.org/api/v1/json/search/tags", format_params(params)
  request = get(search, params=p, proxies=proxies)

  while request.status_code == codes.ok:
    tags, tag_count = request.json()["tags"], 0
    for tag in tags:
      yield tag
      tag_count += 1
    if tag_count < p["per_page"]:
      break

    p["page"] += 1

    request = get(search, params=p, proxies=proxies)

def get_tags(parameters, limit=50, proxies={}):
  params = parameters
  if limit is not None:
    if limit > 0:
      r = tags_requests(params, proxies=proxies)
      for index, tag in enumerate(r, start=1):
        yield tag
        if index >= limit:
          break
  else:
    r = tags_requests(params, proxies=proxies)
    for tag in r:
      yield tag

def get_tag_data(tag, proxies={}):
  url = f"https://derpibooru.org/api/v1/json/tags/{tag}"

  request = get(url, proxies=proxies)

  if request.status_code == codes.ok:
    data = request.json()

    return data["tag"]

def get_user_id_by_name(username, proxies={}):
  url = f"https://derpibooru.org/profiles/{username.replace(' ','+')}"

  request = get(url, proxies=proxies)

  profile_data = request.text
  user_id = profile_data.split("/conversations?with=",1)[-1].split('">',1)[0]
  return user_id

def get_user_data(user_id, proxies={}):
  url = f"https://derpibooru.org/api/v1/json/profiles/{user_id}"

  request = get(url, proxies=proxies)

  if request.status_code == codes.ok:
    data = request.json()

    return data["user"]

def filters_requests(filter_id, params, proxies={}):
  search, p = f"https://derpibooru.org/api/v1/json/filters/{filter_id}", format_params(params)
  request = get(search, params=p, proxies=proxies)

  while request.status_code == codes.ok:
    filters, filter_count = request.json()["filters"], 0
    for filter_item in filters:
      yield filter_item
      filter_count += 1
    if filter_count < p["per_page"]:
      break

    p["page"] += 1

    request = get(search, params=p, proxies=proxies)

def get_filters(filter_id, parameters, limit=50, proxies={}):
  params = parameters

  if limit is not None:
    if limit > 0:
      r = filters_requests(filter_id, params, proxies=proxies)
      for index, filter_item in enumerate(r, start=1):
        yield filter_item
        if index >= limit:
          break
  else:
    r = filters_requests(filter_id, params, proxies=proxies)
    for filter_item in r:
      yield filter_item

def get_filter_data(filter_id, proxies={}):
  url = f"https://derpibooru.org/api/v1/json/filters/{filter_id}"

  request = get(url, proxies=proxies)

  if request.status_code == codes.ok:
    data = request.json()

    return data["filter"]